/* bldhtml.c -- Convert given text file to HTML on stdout and fill in templates.
**
**	usage:		bldhtml <file>
**
** The text file can contain the following constructs:
** 
** Comment:
** 	#<space>...
** 
** Include contents of specified file or text (to end of line) between quotes:
** 	#I[NCLUDE] "filename" ["args"] [args]
** 	#I[NCLUDE] text
** 
** Template for repeated data sequences, e.g., table or list data from file or text:
** 	#T[EMPLATE] "filename"
** 	#T[EMPLATE] text
** 
** Clear TEMPLATE state:
** 	#T[EMPLATE]
** 
** In TEMPLATE state -- Data for current TEMPLATE, one line per replacement field:
** In INPUT state -- Raw data:
** 	text
** 
** End of TEMPLATE definition or replacement text set:
** 	<blankline>
** 
**
** The Template format is:
**	<text><CTRL-A>text<CTRL-B>....
** where the <CTRL-A>'s will be replaced by input text in line order
** and <CTRL-B>'s will be replaced with arguments from the include line.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>

/* possible states */
typedef enum
{
	STATE_INPUT = 0,	/* initial and non-template compiling */
	STATE_TCOMP,		/* compiling a template */
	STATE_TEMPLATE,		/* have template will travel */
	STATE_ERROR			/* error/maximum -- not include in vector table */
}STATE;

#define STATE_MAX	STATE_ERROR	/* maximum possible state value */

/* possible input tokens */
typedef enum
{
	TOKEN_TEXT = 0,		/* text data */ 
	TOKEN_INCFILE,		/* including file */
	TOKEN_INCTEXT,		/* including text */
	TOKEN_TEMFILE,		/* template file data */
	TOKEN_TEMTEXT,		/* template text data */
	TOKEN_TEMEND,		/* reset template state */
	TOKEN_BLANK, 		/* blank line */ 
	TOKEN_ERROR 		/* error or end of the world */ 
}INPUTTOK;

/* pointer to function returning int (STATE) */
typedef STATE (*PFI)();

/* State machine entry. */ 
typedef struct
{
	INPUTTOK input;			/* possible input token */
	PFI vector[STATE_MAX];	/* action vectors depending on current state */
}ENTRY;


/* input parsing function returns TOKEN and relevant string data */
extern INPUTTOK inputparse( FILE *fp, char **retdata );

/* execution functions return next state */
extern STATE passthru( char *data );		/* write data out */
extern STATE tempfill( char *data );		/* process a template */
extern STATE fileread( char *data );		/* read file and write it out */
extern STATE newtemfile( char *data );		/* start a new template from file */
extern STATE temfile( char *data );			/* read template data from file */
extern STATE newtemtext( char *data );		/* start a new template with data */
extern STATE temtext( char *data );			/* continue template with data */
extern STATE cleartem( char *data );		/* clear out template and reset STATE */
extern STATE temend( char *data );			/* reset template to beginning */
extern STATE done( char *data );			/* all done, return ERROR to complete */
extern STATE temdone( char *data );			/* fluxh template and return ERROR */

#define BUFSIZE 1024		/* input buffer size */
#define TBUFSIZE 4*BUFSIZE	/* Template buffer size */

STATE CurrentState = STATE_INPUT;

ENTRY StateMachine[] =
{
  /*  input token	INPUT		TCOMP		TEMPLATE  */
	TOKEN_TEXT,		passthru,	tempfill,	tempfill,
	TOKEN_INCFILE,	fileread,	fileread,	fileread,
	TOKEN_INCTEXT,	passthru,	passthru,	passthru,
	TOKEN_TEMFILE,	newtemfile,	temfile,	newtemfile,
	TOKEN_TEMTEXT,	newtemtext,	temtext,	newtemtext,
	TOKEN_TEMEND,	cleartem,	cleartem,	cleartem,
	TOKEN_BLANK,	passthru,	temend,		temend,
	TOKEN_ERROR,	done,		done,		temdone,
};


void
main( int argc, char **argv )
{
	char *name;
	char *data;
	FILE *src;
	INPUTTOK inptok;

	if( argc != 2 )
	{
		fprintf( stderr, "usage: bldhttp <datafile>\n" );
		exit(1);
	}

	name = argv[1];

	if ( (src = fopen( name, "rt")) == NULL )
	{
		perror( name );
		exit(1);
	}

	fprintf( stderr, "Processing %s...", name );

	/* do the file */
	do
	{
		/* get something to do */
		inptok = inputparse( src, &data );

		/* index into state table and execute based on current state,
		** set next state to be result of execution */
		CurrentState = (*StateMachine[inptok].vector[CurrentState])( data );

	} while( CurrentState != STATE_ERROR ) ;

	fclose( src );
	exit(0);
}


/* this enum and function parse the individual line and
**  figure out if its an file spec, plane text, or nothing there...
** the retdata pointer is mangled to point to the right new thing and
**   hacked and hewed appropriately.
**
*/
typedef enum
{
	INNONE,		/* nothing there.... */
	INTEXT,		/* not quotes */
	INFILE 		/* quotes */
} INTYPE;

INTYPE
text_or_file( char **retdata )
{
	char *cp;
	INTYPE rval = INNONE;

	/* look for a blank space */
	if( cp = strchr( *retdata, ' ' ) )
	{
		/* ok there may be something there.
		** if its a quote then we'll deal with a file, otherwise its text */
		if( cp[1] == '"' )	/* bingo...file */
		{
			/* skip leading quote */
			rval = INFILE;
			*retdata = cp+2;
		}
		else				/* no bingo tonight */
		{
			/* just advance data pointer and be gone */
			rval = INTEXT;
			*retdata = cp+1;
		}
	}

	return( rval );
}


/* input parsing function returns TOKEN and relevant string data */
/*	TOKEN_TEXT = 0,		/* text data */ 
/*	TOKEN_INCFILE,		/* including file */
/*	TOKEN_INCTEXT,		/* including text */
/*	TOKEN_TEMFILE,		/* template file data */
/*	TOKEN_TEMTEXT,		/* template text data */
/*	TOKEN_TEMEND,		/* reset template state */
/*	TOKEN_BLANK 		/* blank line */ 
/*	TOKEN_ERROR 		/* error or end of the world */ 
/*
** Include contents of specified file or text (to end of line) between quotes:
** 	#I[NCLUDE] "filename"
** 	#I[NCLUDE] text
** 
** Template for repeated data sequences, e.g., table or list data from file or text:
** 	#T[EMPLATE] "filename"
** 	#T[EMPLATE] text
** 
** Clear TEMPLATE state:
** 	#T[EMPLATE]
** 
** In TEMPLATE state -- Data for current TEMPLATE, one line per replacement field:
** In INPUT state -- Raw data:
** 	text
** 
** End of TEMPLATE definition or replacement text set:
** 	<blankline>
*/ 

INPUTTOK
inputparse( FILE *fp, char **retdata )
{
	static char Inbuf[BUFSIZE];
	INPUTTOK inptok;
	INTYPE type;

getanother:
	*retdata = Inbuf;	/* default case for text */

	if( fgets( Inbuf, BUFSIZE, fp ) != NULL )
	{
		/* figure out what to do */
		if( Inbuf[0] == '#' )
		{
			/* got a command or comment */
			switch( Inbuf[1] )
			{
				case 'T':
					/* template def, see if its a file or text */
					type = text_or_file( retdata );
					if(      type == INTEXT )	inptok = TOKEN_TEMTEXT;
					else if( type == INFILE ) 	inptok = TOKEN_TEMFILE;
					else if( type == INNONE ) 	inptok = TOKEN_TEMEND;
					else						goto getanother;
				break;

				case 'I':
					/* include def, see if its a file or text */
					type = text_or_file( retdata );
					if(      type == INTEXT )	inptok = TOKEN_INCTEXT;
					else if( type == INFILE ) 	inptok = TOKEN_INCFILE;
					else						goto getanother;
				break;

				case ' ':
				default:
					/* comment, rewind and get another line.... */
					goto getanother;
				break;
			}
		}
		else if( Inbuf[0] == '\n' )
		{
			/* blank line */
			inptok = TOKEN_BLANK;
		}
		else
		{
			/* otherwise its the default TEXT case */
			inptok = TOKEN_TEXT;
		}
	}
	else
	{
		/* error or done */
		*Inbuf = '\0';
		inptok = TOKEN_ERROR;
	}

	return( inptok );

}


/* all execution functions return next state */
/*	STATE_INPUT = 0,	/* initial and non-template compiling */
/*	STATE_TCOMP,		/* compiling a template */
/*	STATE_TEMPLATE,		/* have template will travel */
/*	STATE_ERR			/* error/maximum -- not include in vector table */

STATE
passthru( char *data )			/* write data out */
{
	printf( data );

	return( CurrentState );
}


char TemplateBuf[TBUFSIZE];
char *TemplatePtr = 0;

STATE
tempfill( char *data )			/* process a template */
{
	char *cp;
	int len;

#ifdef DEBUG
	fprintf( stderr, "TEMFILL: " );
	fprintf( stderr, data );
#endif

	if( !TemplatePtr )
	{
		fprintf( stderr, "Sorry man, no template to fill...\n" );
		return( STATE_ERROR );
	}

	/* axe the newline on data in preparation for output */
	if( cp = strchr( data, '\n' ) )
	{
		*cp = '\0';
	}
		
	/* find the next replacement marker */
	if( cp = strchr( TemplatePtr, '\001' ) )
	{
		/* write template up to token and then data */
		len = cp - TemplatePtr;
		printf( "%.*s%s", len, TemplatePtr, data );

		/* advance the template pointer */
		TemplatePtr = cp+1;
	}
	/* if no marker probably a mismatch, deal with it sometime */

	/* note: temend() will write last part of Template buffer */
	return( CurrentState );
}


/* utility funciton to get next word from data string:
**  words are " or space delimited:
**  "one word" another "a third word" last
**   11111111  2222222  333333333333  4444
** First quote should already be removed and return will point to beginning of next word.
** When return points to null you're done.
*/
char*
nextword( char *data )
{
	char *cp;

	/* if you find a quote or a space first, deal with it */
	if( (cp = strchr( data, '"' )) ||
	    (cp = strchr( data, ' ' )) )
	{
		/* skip ahead to the next big thing */
		do
		{
			*cp = '\0';
			++cp;
		} while( *cp && ((*cp == ' ') || (*cp == '"')) || (*cp == '\n') );
	}
	/* otherwise we're probably done */
	else if( (cp = strchr( data, '\n' )) ||
	    	 (cp = strchr( data, '\0' )) )
	{
		*cp = '\0';
	}

	return( cp );
}

/* copy inbuf to outbuf and replace <CTRL-B>'s with words from data string.
** return pointer to next word in data string. do something nice on errors.
*/
char*
parsebuf( char *inbuf, char *outbuf, char *data )
{
	char *ip = inbuf;
	char *dp = data;
	char *next = inbuf;
	char *word;

	while( (next = strchr( ip, '\002' )) )
	{
		word = data;
		data = nextword( data );
		sprintf( outbuf, "%.*s%s", next-ip, ip, word );
		ip = next+1;
		outbuf += strlen( outbuf );
	}
	sprintf( outbuf, "%s", ip );

	return( data );
}


STATE
fileread( char *data )			/* read file and write it out */
{
	char inbuf[BUFSIZE];
	char outbuf[BUFSIZE];
	FILE *src;
	char *next;

	next = data;
	data = nextword( data );

	if ( (src = fopen( next, "rt")) == NULL )
	{
		perror( next );
		return( STATE_ERROR );
	}

	while( fgets( inbuf, BUFSIZE, src ) != NULL )
	{
		if( *inbuf == '#' )	continue;
		data = parsebuf( inbuf, outbuf, data );
		printf( outbuf );
	}

	fclose( src );

	return( CurrentState );
}


/* utility function to open and read template file data */
STATE
readtfile( char *data )
{
	char inbuf[BUFSIZE];
	FILE *src;
	char *next;

	next = data;
	data = nextword( data );

	if ( (src = fopen( next, "rt")) == NULL )
	{
		perror( next );
		return( STATE_ERROR );
	}

	while( fgets( inbuf, BUFSIZE, src ) != NULL )
	{
		if( *inbuf == '#' )	continue;
		data = parsebuf( inbuf, TemplatePtr, data );
		TemplatePtr += strlen( TemplatePtr );
	}

	fclose( src );

	return( STATE_TCOMP );
}


STATE
newtemfile( char *data )		/* start a new template from file */
{
	TemplatePtr = TemplateBuf;

	return( readtfile( data ) );
}


STATE
temfile( char *data )			/* read more template data from file */
{
	if( !TemplatePtr )
	{
		fprintf( stderr, "Not compiling a Template, sorry dude.\n" );
		return( STATE_ERROR );
	}

	return( readtfile( data ) );
}


STATE
newtemtext( char *data )		/* start a new template with data */
{
	TemplatePtr = TemplateBuf;
	strcpy( TemplatePtr, data );
	TemplatePtr += strlen( TemplatePtr );

	return( STATE_TCOMP );
}


STATE
temtext( char *data )			/* continue template with data */
{
	if( !TemplatePtr )
	{
		fprintf( stderr, "Not compiling a Template, sorry dude.\n" );
		return( STATE_ERROR );
	}

	strcpy( TemplatePtr, data );
	TemplatePtr += strlen( TemplatePtr );

	return( STATE_TCOMP );
}


STATE
cleartem( char *data )			/* clear out template and reset STATE */
{
	TemplatePtr = 0;

	return( STATE_INPUT );
}


STATE
temend( char *data )			/* flush template and reset to beginning */
{
#if 1
	if( TemplatePtr && (TemplatePtr != TemplateBuf) )
	{
		printf( TemplatePtr );
	}
#endif

	TemplatePtr = TemplateBuf;

#ifdef DEBUG
	fprintf(stderr, "Fire up TEMPLATE (%s)\n", TemplatePtr );
#endif

	return( STATE_TEMPLATE );
}


STATE
done( char *data )				/* all done, return ERROR to complete */
{
	printf( data );

	return( STATE_ERROR );
}


STATE
temdone( char *data )			/* flush template and return ERROR */
{

	return( STATE_ERROR );
}
